"""
WSGI middleware that makes sure that ``REMOTE_ADDR``
points to the actual client, and not the AWS loadbalancer.

This is needed for proper IP logging, Sentry error reporting,
and potential rate limiting.
"""
import logging

import requests
from netaddr import AddrFormatError, IPNetwork


class UnProxy(object):
    def __init__(self, app):
        self._app = app
        self._allowed_proxy_ips = self._load_allowed_ips()

    def __call__(self, environ, start_response):
        remote_addr = environ.get('REMOTE_ADDR')
        x_forwarded_for = environ.get('HTTP_X_FORWARDED_FOR')

        if x_forwarded_for:
            forwarded_ips = [v.strip() for v in x_forwarded_for.split(',')]

            while self._is_proxy_ip(remote_addr) and forwarded_ips:
                remote_addr = forwarded_ips.pop()

            x_forwarded_for = ", ".join(forwarded_ips)
            _env_set(environ, 'REMOTE_ADDR', remote_addr)
            _env_set(environ, 'HTTP_X_FORWARDED_FOR', x_forwarded_for)
        return self._app(environ, start_response)

    def _is_proxy_ip(self, ip):
        try:
            for addr in self._allowed_proxy_ips:
                if ip in addr:
                    return True
            return False
        except AddrFormatError:
            return False

    def _load_allowed_ips(self):
        """Retrieve the cloudfront ip's from amazon"""

        # Base
        values = [
            '10.0.0.0/8',
            '172.16.0.0/20',
            '192.168.0.0/16',
        ]

        session = requests.session()
        session.mount('http://', requests.adapters.HTTPAdapter(max_retries=5))
        session.mount('https://', requests.adapters.HTTPAdapter(max_retries=5))

        # Cloudfront
        try:
            resp = session.get('https://ip-ranges.amazonaws.com/ip-ranges.json')
        except requests.ConnectionError:
            logger = logging.getLogger(__name__)
            logger.exception("Unable to retrieve AWS ip ranges")
            return []

        if resp.status_code == 200:
            try:
                data = resp.json()
            except ValueError:
                logger = logging.getLogger(__name__)
                logger.exception("Unable to retrieve AWS ip ranges")
                return []

            values.extend([
                x['ip_prefix'] for x in data['prefixes']
                if x['service'] == 'CLOUDFRONT'
            ])
        return [IPNetwork(addr) for addr in values]


def _env_set(environ, key, value):
    """ Sets or deletes the given key in the environ dict. """
    if value:
        environ[key] = value
    elif key in environ:
        del environ[key]



import pytest
import requests_mock
from netaddr import IPNetwork

from wsgi_aws_unproxy import UnProxy


@pytest.fixture
def wsgi_app():
    def app(environ, start_response):
        return "ip={}, ff={}".format(
            environ['REMOTE_ADDR'], environ.get('HTTP_X_FORWARDED_FOR', '')
        )

    json_data = """
    {
        "syncToken": "1490372531",
        "createDate": "2017-03-24-16-22-11",
        "prefixes": [
            {
                "ip_prefix": "13.32.0.0/15",
                "region": "GLOBAL",
                "service": "AMAZON"
            },
            {
                "ip_prefix": "13.54.0.0/15",
                "region": "ap-southeast-2",
                "service": "AMAZON"
            },
            {
                "ip_prefix": "13.32.0.0/15",
                "region": "GLOBAL",
                "service": "CLOUDFRONT"
            }
        ],
        "ipv6_prefixes": [
            {
                "ipv6_prefix": "2400:6500:0:7000::/56",
                "region": "ap-southeast-1",
                "service": "AMAZON"
            }
        ]
    }
    """

    with requests_mock.mock() as rm:
        rm.get(
            'https://ip-ranges.amazonaws.com/ip-ranges.json',
            text=json_data)
        app = UnProxy(app)

    return app


def test_assert_networks(wsgi_app):
    """
    Test that wrapping wsgi object in unproxy doesn't crash.
    """
    assert wsgi_app._allowed_proxy_ips == [
        # Default
        IPNetwork('10.0.0.0/8'),
        IPNetwork('172.16.0.0/20'),
        IPNetwork('192.168.0.0/16'),

        # Extra
        IPNetwork("13.32.0.0/15"),
    ]

    assert not wsgi_app._is_proxy_ip('127.0.0.1')
    assert not wsgi_app._is_proxy_ip('FOOBAR')
    assert wsgi_app._is_proxy_ip('10.0.0.99')
    assert wsgi_app._is_proxy_ip('13.32.0.99')


def test_cloudfront_ip(wsgi_app):
    """
    Should skip first proxy.
    """
    assert wsgi_app({
        'REMOTE_ADDR': '10.0.0.99',
        'HTTP_X_FORWARDED_FOR': '1.2.1.2, 13.32.0.99'
    }, None) == 'ip=1.2.1.2, ff='


def test_non_proxy_ip_multiple(wsgi_app):
    """
    Should return last (is not a proxy).
    """
    assert wsgi_app({
        'REMOTE_ADDR': '10.0.0.99',
        'HTTP_X_FORWARDED_FOR': '1.2.1.2, 1.2.3.3'
    }, None) == 'ip=1.2.3.3, ff=1.2.1.2'


def test_proxy_ip(wsgi_app):
    """
    Remote addr is proxy, returns the forward.
    """
    assert wsgi_app({
        'REMOTE_ADDR': '10.0.0.99',
        'HTTP_X_FORWARDED_FOR': '11.22.33.44'
    }, None) == 'ip=11.22.33.44, ff='


def test_non_proxy_ip(wsgi_app):
    """
    Not a proxy. so REMOTE_ADDR remains unchanged.
    """
    assert wsgi_app({
        'REMOTE_ADDR': '88.88.88.88',
        'HTTP_X_FORWARDED_FOR': '11.22.33.44'
    }, None) == 'ip=88.88.88.88, ff=11.22.33.44'


def test_no_xforwarded_header(wsgi_app):
    """
    No X-Forwarded-For should not crash.
    """
    assert wsgi_app({
        'REMOTE_ADDR': '10.0.0.99'
    }, None) == 'ip=10.0.0.99, ff='


def test_unproxy_bad_json():
    """
    Test that wrapping wsgi object in unproxy doesn't crash
    """
    def app(environ, start_response):
        return environ['REMOTE_ADDR']

    with requests_mock.mock() as rm:
        rm.get('https://ip-ranges.amazonaws.com/ip-ranges.json', text='{NOT_JSON!}')
        app = UnProxy(app)
        assert app._allowed_proxy_ips == []



#!/usr/bin/env python
import re
from setuptools import setup

install_require = [
    'netaddr',
    'requests',
]

docs_require = [
    'sphinx>=1.4.0',
]

tests_require = [
    'requests-mock==1.3.0',
    'pytest-cov>=2.2.0',
    'pytest>=2.8.3',

    # Linting
    'isort==4.2.5',
    'flake8==3.0.3',
    'flake8-blind-except==0.1.1',
    'flake8-debugger==1.4.0',
    'flake8-imports',
]


with open('README.rst') as fh:
    long_description = re.sub('^.. start-no-pypi.*^.. end-no-pypi', '',
                              fh.read(), flags=re.M | re.S)


setup(
    name='wsgi-aws-unproxy',
    version='1.0.0',
    description="Simple wsgi middleware to unproxy AWS",
    long_description=long_description,
    url='https://github.com/labd/wsgi-aws-unproxy',
    author="Lab Digital",
    author_email="opensource@labdigital.nl",
    install_requires=install_require,
    tests_require=tests_require,
    extras_require={
        'docs': docs_require,
        'test': tests_require,
    },
    package_dir={'': 'src'},
    py_modules=['wsgi_aws_unproxy'],
    license='MIT',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
    ],
    zip_safe=False,
)
